//====================================================================================//
//                                                                                    //
//                              PEPS Solution Methods                                 //
//                                                                                    //
//====================================================================================//
//  This File:   solve.C                    Language: C++  (xlC and CC)               //
//  Software:    PEPS 2002                   Advisor: Brigitte.Plateau@imag.fr        //
//  Doc:         solve.tex                                                            //
//====================================================================================//
//  Creation:    19/dec/95                        by: Paulo.Fernandes@imag.fr         //
//  Last Change: 06/nov/02                        by: paulof@inf.pucrs.br             //
//====================================================================================//
#include "peps.h"

//====================================================================================//
//  Power Method                                                                      //
//                                                                                    //
//    This function performs the power method to the infinitesimal generator stored   //
// into the global structure 'CND::cnd' (continuous-time normalized descriptor).      //
//                                                                                    //
//    The explicit input parameters are:                                              //
//    - 'buf'         the initial vector;                                             //
//    - 'p_type'      the type of precondition to be used                             //
//                                                                                    //
//    The implicit input parameters are:                                              //
//    - 'CND::cnd'               the infinitesimal generator                          //
//    - 'PRF::prf.it_type()'     fixed iterations or convergence test                 //
//    - 'PRF::prf.err_type()'    error test type (indiv./accum. - abs./relat.)        //
//    - 'PRF::prf.min_err()'     tolerance for convergence                            //
//    - 'PRF::prf.max_it()'      max. number of iterations                            //
//    - 'PRF::prf.verb()'        verbose mode                                         //
//    - 'PRF::prf.k_size()'      interval between convergence check (if verbose mode) //
//    - 'PRF::prf.p_type()'      Preconditionning type                                //
//                                                                                    //
//    The implicit output parameter is:                                               //
//    - 'VCT::vct'               the resulting vector (solution)                      //
//====================================================================================//
bool Power_CND(VCT & buf)
{
 int   i, k;            // iteration counters
 Timer T;               // timer
 bool  abort = false;   // a flag to abort the iteration after the initial test
 bool  conv;            // the precision for convergence was reached
 rp    resid;           // the residue of the vector
 rp    acc;             // the sum of all elements of the vector
 rp    largest;         // the largest element of the vector
 int   lwhere;          // the position of the largest element of the vector
 rp    smallest;        // the smallest element of the vector
 int   swhere;          // the position of the smallest element of the vector

 bool  test_error       = (PRF::prf.it_type() != fix_it); // verify convergence
 e_typ error_test_type  = PRF::prf.err_type();            // error verification type
 int   limit            = PRF::prf.max_it();              // max. number of iterations
 rp    tol              = PRF::prf.min_err();             // tolerance for convergence
 bool  verbose          = PRF::prf.verb();                // verbose mode
 int   big_step         = (PRF::prf.k_size() / 2);        // half interval between info
 precond_typ p_type     = PRF::prf.p_type();              // Preconditionning type

 cout.setf(ios::scientific);
 cout.precision(peps_precision);
 VCT::vct.alloc_static();

 cout << "Iteration 0: "; cout.flush();
 T.clear(); T.start();
 VCT::vct.mult_CND(buf);

 T.stop();
 resid = VCT::vct.norm_inf();
 i = 1;
 conv = (resid < tol);

 VCT::vct.add(buf);

 acc = VCT::vct.sum();
 if ( (acc < one - round_zero) || (acc > one + round_zero) )            // initial test
   if  ( (acc < - PRF::prf.min_err()) || (acc > PRF::prf.min_err()) ) {
     cout << "\nWARNING: sum of residual vector = " << acc << " (it should be zero)"
          << "\nIteration 0: ";
   }
   else {
     cout << "\n*******************************************************************"
          << "\n***       the descriptor is not a infinitesimal generator       ***"
          << "\n*******************************************************************"
          << "\n                   Aborting the iterations"
          << "\n Too large sum of residual vector = " << acc  << " (it should be zero)\n";
     abort = true;
   }
 if (!abort) {
   if (p_type == no_precond)
     CND::cnd.add_identity();
   T.start();
   //-----------------------------------------------------------------------------------
   while ((i < limit) && !conv) {
     if (verbose) {
       VCT::vct.info(acc, largest, lwhere, smallest, swhere);
       cout << setiosflags(ios::scientific)
            << "largest: "  << largest  << " (" << lwhere << ") "
            << "smallest: " << smallest << " (" << swhere << ") ";
       if ( (acc < one - round_zero) || (acc > one + round_zero) )
         cout << " - sum: " << acc << " !!!\n";
       else
         cout << "\n";
     }
     for (k=0; (k<big_step && i<limit); k++) {
       cout << "\r                                      "
            << "\rIteration " << i << ": "; cout.flush();
       //======================
       switch (p_type) {
         case no_precond : buf.mult_CND(VCT::vct); 
	                                                   break;
         case diag       : buf.pre_diag(VCT::vct); 
                           buf.norm_sum(VCT::vct);  break;
         case plyn       : buf.pre_plyn(VCT::vct); 
                           buf.norm_sum(VCT::vct);  break;
       }
       //======================
       i++;
       //********************************************************
       //****************** Operation included by Billy's request
        buf.normalize();
       //********************************************************
       if (test_error)
         if (conv = VCT::vct.test(buf, error_test_type, tol)) {
           VCT::vct.copy(buf);
           break;
         }
       if (i < limit) {
         cout << "\r                                      "
              << "\rIteration " << i << ": "; cout.flush();
         //======================
         switch (p_type) {
           case no_precond : VCT::vct.mult_CND(buf);
                                                      break;
           case diag       : VCT::vct.pre_diag(buf); 
                             VCT::vct.norm_sum(buf);  break;
           case plyn       : VCT::vct.pre_plyn(buf); 
                             VCT::vct.norm_sum(buf);  break;
         }
         //======================
         i++;
         //********************************************************
         //****************** Operation included by Billy's request
          VCT::vct.normalize();
         //********************************************************
         if (test_error)
           if (conv = VCT::vct.test(buf, error_test_type, tol))
             break;
       }
       else {
         VCT::vct.copy(buf);
       }
     }
   }
   //-----------------------------------------------------------------------------------
   T.stop();

   if (p_type == no_precond)
     CND::cnd.minus_identity();
   if (PRF::prf.it_type() == cnv_it) {        // Convergence verification
     buf.mult_CND(VCT::vct);                  //   One extra multiplication is made
     resid = buf.norm_inf();                  //     and the largest element is
   }                                          //     considered as the residue
   Notify_Time_Spend(T, "Power solution", i, resid);
   VCT::vct.write();
   Log_Out_Solution(T, i, Power, p_type, resid);
 }

 VCT::vct.desalloc_static();
 return(resid < PRF::prf.min_err());
}


//===========================================================================================//
//  Standard Uniformization Method  - Transient solution                                     //
//                                                                                           //
//    This function performs the uniformization method to the infinitesimal generator stored //
// into the global structure 'CND::cnd' (continuous-time normalized descriptor).   	       //
// 	The Standard Uniformization is based on  the reduction of a Continuous-Time structure   //
// to a continuous-Time Normalized subordinaded to a Poisson Process.  		                //
//                                                                                           //
//    The explicit input parameters are:                                                     //
//    - 'buf'         the initial vector;                                                    //
//    - 'p_type'      the type of precondition to be used                                    //
//    - 'tm'	      the time                                                                 //
//                                                                                           //
//    The implicit input parameters are:                                                     //
//    - 'CND::cnd'               the infinitesimal generator                                 //
//    - 'PRF::prf.it_type()'     fixed iterations or convergence test                        //
//    - 'PRF::prf.err_type()'    error test type (indiv./accum. - abs./relat.)               //
//    - 'PRF::prf.min_err()'     tolerance for convergence                                   //
//    - 'PRF::prf.max_it()'      max. number of iterations                                   //
//    - 'PRF::prf.verb()'        verbose mode                                                //
//    - 'PRF::prf.k_size()'      interval between convergence check (if verbose mode)        //
//                                                                                           //
//    The implicit output parameter is:                                                      //
//    - 'VCT::vct'               the resulting vector (solution)                             //
//===========================================================================================//

bool Uniformization_CND(VCT & buf, rp _time)
{
 int   i, j, k;         // iteration counters
 Timer T;               // timer
 bool  abort = false;   // a flag to abort the iteration after the initial test
 bool  conv;            // the precision for convergence was reached
 rp    resid;           // the residue of the vector
 rp    acc;             // the sum of all elements of the vector
 rp    largest;         // the largest element of the vector
 int   lwhere;          // the position of the largest element of the vector
 rp    smallest;        // the smallest element of the vector
 int   swhere;          // the position of the smallest element of the vector
 rp    time;            // the time
 rp    bigest;          // the bigest diagonal element 
 rp    bigest_t;        // the bigest diagonal element times time
 rp    prod;
 rp    mbig;
 rp    sum;
 rp    sigma;
 int   maxit;           // max iterations in external loop
 VCT   aux;             // auxiliar vector

 bool  test_error       = (PRF::prf.it_type() != fix_it); // verify convergence
 e_typ error_test_type  = PRF::prf.err_type();            // error verification type
 rp    tol              = PRF::prf.min_err();             // tolerance for convergence
 bool  verbose          = PRF::prf.verb();                // verbose mode

 cout.setf(ios::scientific);
 cout.precision(peps_precision);
 VCT::vct.alloc_static();
 aux.create(VCT::vct.size(), VCT::vct.real_size());
 aux.get_CND();

 time = _time;
 bigest = 1/CND::cnd.n_factor();

 cout << "bigest: " << bigest << endl;

 bigest_t = bigest * time;
 maxit = 1;
 if (bigest_t > 100){ // number of external iterations
   maxit = 1 + (int)(bigest_t/100);
   time = time/maxit;
 }  

 bigest_t = bigest * time;
 prod = exp(-bigest_t);
 mbig = 0;
 sum = prod;
 sigma = 1.0;
 while(!(sum > (1.0-tol))){ // number of internal iterations
   mbig++;
   sigma = sigma * (bigest_t/mbig);
   sum += sigma * prod;
 } 

 cout << "Iteration 0/" << (int)(maxit*mbig) << ": "; cout.flush();
 VCT::vct.mult_CND(buf);

 resid = VCT::vct.norm_inf();
 i = 1;
 conv = (resid < tol);

 VCT::vct.add(buf);

 acc = VCT::vct.sum();
 if ( (acc < one - round_zero) || (acc > one + round_zero) )            // initial test
   if  ( (acc < - PRF::prf.min_err()) || (acc > PRF::prf.min_err()) ) {
     cout << "\nWARNING: sum of residual vector = " << acc << " (it should be zero)"
          << "\nIteration 0: ";
   }
   else {
     cout << "\n*******************************************************************"
          << "\n***       the descriptor is not a infinitesimal generator       ***"
          << "\n*******************************************************************"
          << "\n                   Aborting the iterations"
          << "\n Too large sum of residual vector = " << acc  << " (it should be zero)\n";
     abort = true;
   }

 if(!abort){
   CND::cnd.add_identity();
   T.clear(); T.start();
   for(i=0; i<maxit; i++){ // external loop
     VCT::vct.copy(buf);
     if (verbose) {
       VCT::vct.info(acc, largest, lwhere, smallest, swhere);
       cout << setiosflags(ios::scientific)
            << "largest: "  << largest  << " (" << lwhere << ") "
            << "smallest: " << smallest << " (" << swhere << ") ";
       if ( (acc < one - round_zero) || (acc > one + round_zero) )
         cout << " - sum: " << acc << " !!!\n";
       else
         cout << "\n";
     }
     for(int k=1; k<=mbig; k++){ // internal loop
       cout << "\r                                                     "
            << "\rIteration " << (int)((i*mbig)+k) << "/" << (int)(maxit*mbig) << ": "; cout.flush();
       aux.mult_CND(VCT::vct);
       VCT::vct.scale_copy(aux,(bigest_t/k));
       buf.add(VCT::vct);
     }
     buf.scale(prod);
     buf.normalize();
   }
   T.stop(); 
   CND::cnd.minus_identity();
   time = time*maxit;

   if (PRF::prf.it_type() == cnv_it) {        // Convergence verification
     VCT::vct.mult_CND(buf);                  //   One extra multiplication is made
     resid = VCT::vct.norm_inf();             //     and the largest element is
   }                                          //     considered as the residue

   Notify_Time_Spend(T, "Uniformization solution", (int)(maxit*mbig), resid);
   VCT::vct.copy(buf);
   VCT::vct.write();
   Log_Out_Solution(T, (int)(maxit*mbig), Uniformization, no_precond, resid);
 }
 VCT::vct.desalloc_static();
 return(resid < PRF::prf.min_err());
}


